//-------------------------------------------------------------------------------------------------------
//
// IH_CON  -  Main
//
// Einige Programmteile wurden von anderen Projekten bernommen:
//  - I2C: Peter Fleury
//  - MCP2515/CAN: Fabian Greif (www.kreatives-chaos.de)
//
// Funktionen:
//	- Auswertung der 14 digitalen Eingnge: 
//	  1x kurz, 2x kurz, 1x lang, Status und wiederkehrend bei gedrckt
//	- CAN Kommunikation
//	- internes EEPROM schreiben und lesen
//	- mehrere Timer (krzeste ist 10ms)
//	- Uhrzeit wird sekndlich ber CAN empfangen, es wird nur im 10ms Takt selber gezhlt
//	- Fehlerauswertung und -anzeige
//	- Ausgnge knnen ber CAN geschalten werden
//	- 3 Ausgnge knnen als PWM zur Dimmeransteuerung verwendet werden
//	- Timer 0: 0,01ms - Dimmer PWM
//	- Timer 1: 10ms - Takt und DE Auswertung
//
//-------------------------------------------------------------------------------------------------------


#include "Config.h"

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/wdt.h>
#include <avr/sleep.h>

#include <util/delay.h>
#include <stdio.h>
#include <stdlib.h>
#include <inttypes.h>

#include "MCP2515.h"
#include "Error.h"
#include "SPI.h"
#include "EEPROM.h"
#include "DE.h"
#include "Stromaus.h"
#include "Funktionen.h"
#include "Module.h"


//-------------------------------------------------------------------------------------------------------
// Funktionsprototypen
ISR(TIMER1_OVF_vect);			// Timer 1: 10ms - Dimmer PWM, Takt und DE-Auswertung
ISR(TIMER1_COMPA_vect);			// Timer 1: Dimmer 0 PWM
ISR(TIMER1_COMPB_vect);			// Timer 1: Dimmer 1 PWM


//-------------------------------------------------------------------------------------------------------
// Hauptprogramm
int main(void)
{
	// Reset Status Register zurck setzen nach Watchdog Reset
	MCUSR &= ~(1<<WDRF);
	wdt_disable();
		
	
	// Statusled
	SET_OUTPUT(LED_G);
	SET_OUTPUT(LED_R);
	
	// 24V Ausgnge
	SET_OUTPUT(DA1);
	SET_OUTPUT(DA2);
	SET_OUTPUT(DA3);
	SET_OUTPUT(DA4);
	SET_OUTPUT(DA5);
	SET_OUTPUT(DA6);
	SET_OUTPUT(DA7);
	SET_OUTPUT(DA8);

	RESET(DA1);
	RESET(DA2);
	RESET(DA3);
	RESET(DA4);
	RESET(DA5);
	RESET(DA6);
	RESET(DA7);
	RESET(DA8);

	
	// Variablen
	uint8_t i = 0;
	
	uint8_t EEPROM_Daten[4];	// zum Beschreiben und Auslesen des EEPROMs
	uint8_t EEPROM_Laenge = 0;	// Anzahl der Datenbytes
	uint8_t adresse_h;			// EEPROM Speicheradresse High
	uint16_t adresse_l;			// EEPROM Speicheradresse Low
	//uint8_t Schlafmodus = 0;	// Zhler fr Schlafmodus
	uint8_t Sekunde_alt = 0;	// Merker fr Led Pulsschlag
	uint8_t Minute_alt = 0;
	uint8_t Uhrzeit_empfangen = 0;
	

	
	// Timer 1: 10ms - Takt, DE Auswertung und Dimmer
	TCCR1A |= (1<<WGM11);								// Fast PWM Mode mit ICR1 als Top
	TCCR1B |= (1<<WGM12) | (1<<WGM13) | (1<<CS11);		// Fast PWM und Prescale=8
	ICR1 = Timer1_Top;									// 100 Hz
	
	#if Dimmer_0
		#if Dimmer_1
			TIMSK1 |= (1<<TOIE1) | (1<<OCIE1A) | (1<<OCIE1B);	// Overflow und Capture Interrupt A und B
		#else
			TIMSK1 |= (1<<TOIE1) | (1<<OCIE1A);					// Overflow und Capture Interrupt A
		#endif
	#else
		TIMSK1 |= (1<<TOIE1);									// Overflow Interrupt
	#endif


	
	// Schnittstellen initialisieren
	F_SPI_init();
	
	// Fehleranzeige initialisieren
	F_Error_init();
	
	// DE Auswertung initialisieren
	F_DE_init();
	
	// PWM Dimmer initialisieren
	F_Dimmer_init();
	
	
	// MCP2515 initialisierung
	if(F_MCP2515_init() == 0)
	{
		F_Error(11, 0, 0, 0, 0, 0);  // Fehlertext: CAN Receiver initialisieren nicht mglich
	}

	// CAN Nachricht erzeugen
	tCAN message_rec;			// zum Empfangen
	
	
	// Interrupts einschalten
	sei();
	
	
	_delay_ms(CAN_Adresse * 10);


	

	// Endlosschleife
	while(1)
	{
		// wird einmal beim Start nach dem Empfangen der ersten Uhrzeit ausgefhrt
		if(Uhrzeit_empfangen == 1)
		{
			Uhrzeit_empfangen = 2;
			
			// Rollos initialisieren
			F_Rollo_init();
			
			// Initialisierungsfunktionen
			F_Init();
		}
		
		
		// Led Pulsschlag
		if(Sekunde != Sekunde_alt)
		{
			Sekunde_alt = Sekunde;
			
			if(Error_Stand == 0)  // kein Fehler zum Anzeigen
			{
				if((Sekunde % 2) == 0)
					SET(LED_G);
				else
					RESET(LED_G);
			}
		}	
		
		// mintliche Ausfhrung
		if(Minute != Minute_alt)
		{
			Minute_alt = Minute;
			F_Finster_Hell();
		}
		
		// Fehlerausgabe
		F_Error_Ausgabe();
		
		// DE Auswertung
		F_DE_betaetigt();
		
		// Timer
		F_Timer();
		
		// Automatikfunktionen
		F_Auto();
		
		// Rolloansteuerung
		F_Rollo();

		// CAN Puffer abarbeiten
		F_CAN_Puffer();
		
		
		// CAN Telegramm empfangen
		if(MCP2515_tel_empfangen > 0)  // CAN Telegramm empfangen
		{
			F_MCP2515_tel_empfangen(&message_rec);
			MCP2515_tel_empfangen--;

			if(message_rec.data[0] == CAN_Adresse || message_rec.data[0] == 99)  // Prfen, ob die Adresse passt
			{
				switch(message_rec.data[1])  // Telegrammtyp checken
				{
					case 10:	// Ausgang schalten
								// Rckmeldung schicken
								F_CAN_senden(	message_rec.id, 75, message_rec.data[7], 0, 0, 0, 0, 0);

								if(message_rec.data[2] >= 1 && message_rec.data[2] <= 8)
								{
									if(message_rec.data[3] == 0)  // ausschalten
									{ 
										F_Ausgang_ausschalten(message_rec.data[2]);
									}
									else if(message_rec.dtat[3] == 200)	// toggeln
									{
										F_Ausgang_toggeln(message_rec.data[2]);
									}
									else if(message_rec.data[3] > 0 && message_rec.data[3] <= 100) // einschalten
									{
										F_Ausgang_einschalten(message_rec.data[2], message_rec.data[3]);
									}
								}
								else
								{
									// Fehlermeldung: Telegrammtyp 10: falsche Ausgangsnummer
									F_Error(16, message_rec.id, message_rec.data[2], 0, 0, 0);
								}

								break;
										
					
					case 16:	// Rollo ansteuern
								// Rckmeldung schicken
								F_CAN_senden(	message_rec.id, 75, message_rec.data[7], 0, 0, 0, 0, 0);

								switch(message_rec.data[2])
								{
									case 1: // Rollo 0
											#if Rollo_0
											F_Rollo_Pos(0, message_rec.data[3]);
											#endif
											break;

									case 2: // Rollo 1
											#if Rollo_1
											F_Rollo_Pos(1, message_rec.data[3]);
											#endif
											break;

									default: break;
								}

								break;


					case 20:	// LED-Modul Status
								Led_Modus[message_rec.id - 60] = message_rec.data[2];
								Led_Info_1[message_rec.id - 60] = message_rec.data[3];
								Led_Info_2[message_rec.id - 60] = message_rec.data[4];
								Led_Info_3[message_rec.id - 60] = message_rec.data[5];
								Led_Info_4[message_rec.id - 60] = message_rec.data[6];
								break;


					case 21:	// CON Status
								break;


					case 30:	// EEPROM Beschreiben
								if(message_rec.data[4] >= 1 && message_rec.data[4] <= 3)
								{
									for(i=0; i < message_rec.data[4]; i++)
									{
										F_EEPROM_schreiben(message_rec.data[2], message_rec.data[3] + i, message_rec.data[5 + i]);
									}
								}
								else
								{
									// Fehlertext: Empf. Telegramm 30: falscher Wert Datenbyte 4
									F_Error(12, message_rec.id, message_rec.data[4], 0, 0, 0);	
								}
								break;


					case 31:	// EEPROM Anforderung Auslesen
								adresse_h = message_rec.data[2];
								adresse_l = message_rec.data[3];
								EEPROM_Laenge = 0;
								EEPROM_Daten[0] = 0;
								EEPROM_Daten[1] = 0;
								EEPROM_Daten[2] = 0;
								EEPROM_Daten[3] = 0;

								if(message_rec.data[4] >= 1 && message_rec.data[4] <=4)
								{
									for(i=0; i < message_rec.data[4]; i++)
									{
										EEPROM_Daten[i] = F_EEPROM_lesen(adresse_h, adresse_l);

										adresse_l++;
										if(adresse_l == 256)
										{
											adresse_l = 0;
											adresse_h++;

											if(adresse_h >= 8)
											{
												// Fehlertext: Empf. Telegramm 31: EEPORM Adresse High zu hoch
												F_Error(14, message_rec.id, message_rec.data[2], message_rec.data[3], message_rec.data[4], 0);	
												
												i = message_rec.data[4];
												EEPROM_Laenge = 0;
												EEPROM_Daten[0] = 0;
												EEPROM_Daten[1] = 0;
												EEPROM_Daten[2] = 0;
												EEPROM_Daten[3] = 0;
											}
										}

										EEPROM_Laenge++;
									}
							
									// Antwort erstellen
									F_CAN_senden(	message_rec.id, 
											32, 
											message_rec.data[2],
											message_rec.data[3],
											EEPROM_Daten[0],
											EEPROM_Daten[1],
											EEPROM_Daten[2],
											EEPROM_Daten[3]);
								}
								
								else
								{
									// Fehlertext: Empf. Telegramm 31: falscher Wert Datenbyte 4
									F_Error(13, message_rec.id, message_rec.data[4], 0, 0, 0);	
								}
								break;


					case 40:	// Sensorwerte
								switch(message_rec.data[2])
								{
									case 1: // Temperatur										
											if(message_rec.data[3] >= 100)	// Minusgrad
											{
												Temp_Au = (message_rec.data[4] + ((message_rec.data[3]-100) / 10)) * -1;
											}
											
											else  // Plusgrade
												Temp_Au = message_rec.data[4] + (message_rec.data[3] / 10);
											
											break;
									
									case 2:	// relative Luftfeuchte
											Feuchte_Au = message_rec.data[3];
											break;
									
									case 3: // Helligkeit
											Helligkeit = message_rec.data[3] | (message_rec.data[4] << 8);
											break;
									
									case 4: // Luftdruck
											Luftdruck = message_rec.data[3] + 900;
											break;
									
									default: break;
								}
								
								switch(message_rec.data[5])
								{
									case 1: // Temperatur
											if(message_rec.data[6] >= 100)	// Minusgrad
											{
												Temp_Au = (message_rec.data[7] + ((message_rec.data[6]-100) / 10)) * -1;
											}
									
											else  // Plusgrade
												Temp_Au = message_rec.data[7] + (message_rec.data[6] / 10);
									
											break;
									
									case 2:	// relative Luftfeuchte
											Feuchte_Au = message_rec.data[6];
											break;
									
									case 3: // Helligkeit
											Helligkeit = message_rec.data[6] | (message_rec.data[7] << 8);
											break;
									
									case 4: // Luftdruck
											Luftdruck = message_rec.data[6] + 900;
											break;
									
									default: break;
								}
								break;
					
					
					case 70:	// Steuerbefehle
								// Rckmeldung schicken
								F_CAN_senden(	message_rec.id, 75, message_rec.data[7], 0, 0, 0, 0, 0);

								switch(message_rec.data[2])
								{
									case 1:	// Softwarereset
											cli();
											wdt_enable(WDTO_15MS);
											while(1);
											break;
						
									case 2:	// Strom-Aus Meldung
											F_Stromaus();											
											break;
						
									case 3: // alle Fehler zurck setzen
											F_Error_init();
											break;
						
									default: break;
								}
								break;
					
					
					case 80:	// aktuelle Zeit
								Stunde = message_rec.data[2];
								Minute = message_rec.data[3];
								Sekunde = message_rec.data[4];
								Ms10 = 0;
								Tag = message_rec.data[5];
								Wochentag = message_rec.data[6] & 0x0F;
								Monat = (message_rec.data[6] & 0xF0) >> 4;
								Jahr = message_rec.data[7];
								
								if(Uhrzeit_empfangen == 0)
									Uhrzeit_empfangen = 1;
								break;
								
					
					case 90:	// Fehlertelegramm															
								break;
					

					default:	break;
				}
			}
			
			
			else if(message_rec.data[0] >= 60 && message_rec.data[0] <= 67)  // LED Module
			{
				if(message_rec.data[1] == 15)	// LED Modul Ansteuerung
				{
					Led_Modus[message_rec.id - 60] = message_rec.data[2];		// Betriebsmodus (0=aus, 1=einzelne Ausgnge, 2=Farbe, 3=auto Farbwechsel)
					Led_Info_1[message_rec.id - 60] = message_rec.data[3];		// Modus 1: DA1 Sollwert, Modus 2: Farbe, Modus 3: Geschwindigkeit
					Led_Info_2[message_rec.id - 60] = message_rec.data[4];		// Modus 1: DA2 Sollwert, Modus 2&3: Dimmwert
					Led_Info_3[message_rec.id - 60] = message_rec.data[5];		// Modus 1: DA3 Sollwert
					Led_Info_4[message_rec.id - 60] = message_rec.data[6];		// Modus 1: DA4 Sollwert
				}
			}
		}

		
		/*// Schlafmodus
		Schlafmodus++;

		if(Schlafmodus >= 5)
		{
			Schlafmodus = 0;
			 
			set_sleep_mode(SLEEP_MODE_IDLE);  // Schlafmodus definieren
			sleep_mode();  // Schlafen einschalten
		}*/
	}
}



//-------------------------------------------------------------------------------------------------------
// Timer 1: 10ms - Dimmer, Takt und DE Auswertung

ISR(TIMER1_OVF_vect)
{
	// 10ms Takt
	Ms10++;


	// DE Auswertung

	static uint16_t ct0, ct1;
	static uint16_t rpt[14];
	static uint16_t DE_k_Erkannt;
	static uint16_t rpt_kurz[14];
	static uint16_t DE_Status_alt;  // Entprellter und invertierter Status des vorherigen Durchgangs
	static uint16_t DE_k_Zeit;  // Wartezeit fr 1 oder 2x kurz einschalten
	static uint8_t DE_Flanke_fall_k;	// Merker fr 2x kurz bettigung
	uint16_t i, pins;
	uint8_t x;
	
	// alle DE auf eine Varible zusammenfgen
	pins = PINC & 0xF0;
	pins = pins << 4;
	pins |= PINA;
	pins = pins << 2;
	pins |= PINB & 0x03;

	DE_Status_alt = DE_Status;

	i = DE_Status^ ~pins;      // nderung bei DE?
	ct0 = ~(ct0 & i);          // ct0 erhhen oder zurcksetzen
	ct1 = ct0 ^ (ct1 & i);     // ct1 erhhen oder zurcksetzen
	i &= ct0 & ct1;            // erhhen bis berlauf
	DE_Status ^= i;            // danach DE_Status ndern


	DE_Zwei |= DE_k_Zeit & ~DE_Status & DE_Flanke_fall_k & DE_Status_alt;  // 0->1: zweite Bettigung erkannt
	DE_k_Erkannt = DE_Status & i;     // 0->1: kurze Bettigung erkannt

	DE_k_Zeit |= DE_k_Erkannt & ~DE_Status_alt & ~DE_Flanke_fall;  // Zeitverzgerung fr 1x/2x starten

	DE_Flanke_fall |= (DE_Status ^ DE_Status_alt) & ~DE_Status;  // 0->1: Negative Flanke erkannt
	DE_Flanke_fall_k |= (DE_Status ^ DE_Status_alt) & ~DE_Status & DE_k_Zeit;	// Merker fr 2x kurz bettigung
	DE_Flanke_steig |= (DE_Status ^ DE_Status_alt) & DE_Status; // 1->0: Positive Flanke erkannt
	

	for(x=0; x<=13; x++)			// Wiederholfunktion
	{
		if((DE_Status & (1<<x)) == 0)
			rpt[x] = DE_Zeit_lang;

		if(--rpt[x] == 0)
		{
			rpt[x] = DE_Zeit_repeat;
			DE_Repeat |= DE_Status & (1<<x);
			
			// welche Eingnge nutzen die Repeatfunktion:
			if(Repeat_Maske)
				DE_Erkannt &= ~(1<<x);
		}
	}
	
	for(x=0; x<=13; x++)			// 1x oder 2x kurz
	{
		rpt_kurz[x]--;

		if(rpt_kurz[x] == 0)
		{
			DE_Zwei &= ~(1<<x);  // DE_Zwei zurck setzen
		}

		if((DE_k_Zeit & (1<<x)) == 0)
			rpt_kurz[x] = DE_Zeit_kurz;  // muss kleiner als bei der Wiederholfunktion sein

		if(rpt_kurz[x] == 1)
		{
			DE_Flanke_fall_k &= ~(1<<x);  // DE_Flanke zurck gesetzt
			DE_Erkannt |= 1<<x;  // 0->1 DE Bettigung erkannt
			DE_k_Zeit &= ~(1<<x);  // 1x/2x Zeitverzgerung ausschalten
		}
	}

	
	// PWM starten
	#if Dimmer_0
	if(Dimmer_Ist[0] > Dimmer_Min[0])	// Dimmer 0
		SET(Dimmer_0_DA);
		
	else
		Dimmer_Status[0] = 0;
	#endif
	
	#if Dimmer_1
		#if Sonderfall_SZ
		if(Sonderfall_SZ_links == 1)
			SET(Dimmer_1_DA);
		if(Sonderfall_SZ_rechts == 1)
			SET(Dimmer_1_DA2);
		#else
		if(Dimmer_Ist[1] > Dimmer_Min[1])	// Dimmer 1
			SET(Dimmer_1_DA);
		
		else
			Dimmer_Status[1] = 0;
		#endif
	#endif


	// Dimmer 0 Istwert anpassen  Fade
	#if Dimmer_0
	if((Dimmer_Ist[0] < Dimmer_Soll_MM[0]) && (Dimmer_Ist[0] < Dimmer_Min[0]) && (Dimmer_Soll_MM[0] != 0))
	{
		Dimmer_Ist[0] = Dimmer_Min[0];
		OCR1A = Dimmer_Log[Dimmer_Ist[0]];	// neuen PWM Wert setzen
		Dimmer_Status[0] = 10;
	}

	else if((Dimmer_Ist[0] < Dimmer_Soll_MM[0])  && (Dimmer_Ist[0] >= Dimmer_Min[0]) && (Dimmer_Soll_MM[0] != 0))
	{
		Dimmer_Ist[0] ++;
		OCR1A = Dimmer_Log[Dimmer_Ist[0]];	// neuen PWM Wert setzen
		Dimmer_Status[0] = 10;
	}
	
	else if((Dimmer_Ist[0] > Dimmer_Soll_MM[0]) && (Dimmer_Ist[0] > Dimmer_Min[0]))
 	{
		Dimmer_Ist[0] --;
		OCR1A = Dimmer_Log[Dimmer_Ist[0]]; 	// neuen PWM Wert setzen
		Dimmer_Status[0] = 11;
	}
	
	else if((Dimmer_Ist[0] == Dimmer_Soll_MM[0]) && (Dimmer_Soll[0] != 0))
	{
		Dimmer_Status[0] = 1;
	}
	
	else if((Dimmer_Ist[0] <= Dimmer_Min[0]) && (Dimmer_Soll_MM[0] == 0))
	{
		Dimmer_Ist[0] = 0;
		Dimmer_Status[0] = 0;
	}
	#endif
	
	
	// Dimmer 1 Istwert anpassen  Fade
	#if Dimmer_1
		#if (Sonderfall_SZ == 0)
		if((Dimmer_Ist[1] < Dimmer_Soll_MM[1]) && (Dimmer_Ist[1] < Dimmer_Min[1]) && (Dimmer_Soll_MM[1] != 0))
		{
			Dimmer_Ist[1] = Dimmer_Min[1];
			OCR1B = Dimmer_Log[Dimmer_Ist[1]];	// neuen PWM Wert setzen
			Dimmer_Status[1] = 10;
		}

		else if((Dimmer_Ist[1] < Dimmer_Soll_MM[1])  && (Dimmer_Ist[1] >= Dimmer_Min[1]) && (Dimmer_Soll_MM[1] != 0))
		{
			Dimmer_Ist[1] ++;
			OCR1B = Dimmer_Log[Dimmer_Ist[1]];	// neuen PWM Wert setzen
			Dimmer_Status[1] = 10;
		}
	
		else if((Dimmer_Ist[1] > Dimmer_Soll_MM[1]) && (Dimmer_Ist[1] > Dimmer_Min[1]))
 		{
			Dimmer_Ist[1] --;
			OCR1B = Dimmer_Log[Dimmer_Ist[1]]; 	// neuen PWM Wert setzen
			Dimmer_Status[1] = 11;
		}
	
		else if((Dimmer_Ist[1] == Dimmer_Soll_MM[1]) && (Dimmer_Soll[1] != 0))
		{
			Dimmer_Status[1] = 1;
		}
	
		else if((Dimmer_Ist[1] <= Dimmer_Min[1]) && (Dimmer_Soll_MM[1] == 0))
		{
			Dimmer_Ist[1] = 0;
			Dimmer_Status[1] = 0;
		}
		#endif
	#endif
	
}



//-------------------------------------------------------------------------------------------------------
// Timer 1: 10ms - Dimmer 0, Compare Match fr PWM

ISR(TIMER1_COMPA_vect)
{
	#if Dimmer_0
	RESET(Dimmer_0_DA);
	#endif
	
	
}


//-------------------------------------------------------------------------------------------------------
// Timer 1: 10ms - Dimmer 1, Compare Match fr PWM

ISR(TIMER1_COMPB_vect)
{
	#if Dimmer_1
		#if Sonderfall_SZ
		RESET(Dimmer_1_DA);
		RESET(Dimmer_1_DA2);
		#else
		RESET(Dimmer_1_DA);
		#endif
	#endif
}